/*
 * Lamer Tetris - moja implementacja kultowej gry
 * new version: 06-VIII-2021
 * Przemysław R. Pietraszczyk
 * Licencja: Public Domain
 * Powered by: Geany
*/

#include <stdio.h>   // printf
#include <stdlib.h>  // exit
#include <string.h>
#include <unistd.h>  // sleep  access
#include <time.h>
//#include <errno.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_ttf.h>

#if WINDOWS
#include <windows.h>
#include <filesystem>
#include <iostream>
#include <fstream>
#include <sys/stat.h>
//#include "resource.h"
#endif

#include "field-szlam.h"

#define SIZEFONT 35
#define SCOREFONT 28

SDL_Window  *gWindow = NULL;
SDL_Renderer * renderer = NULL;
SDL_Texture * tex_screen = NULL;
SDL_Surface * game_screen = NULL;
SDL_Surface * wallpaper = NULL;
TTF_Font* font_panel = NULL, * font_score = NULL;

int hiscore=0, score = 0;

Uint32 getpixel(SDL_Surface *surface, int x, int y)
{
	int bpp = surface->format->BytesPerPixel;
	/* Here p is the address to the pixel we want to retrieve */
	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

	switch(bpp) {
		case 1:
			return *p;
		case 2:
			return *(Uint16 *)p;
		case 3:
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
				return p[0] << 16 | p[1] << 8 | p[2];
			else
				return p[0] | p[1] << 8 | p[2] << 16;
		case 4:
			return *(Uint32 *)p;
		default:
			return 0;
	}
}



void putpixel(SDL_Surface *surface, int x, int y, Uint8 R, Uint8 G, Uint8 B)
{
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x *
               surface->format->BytesPerPixel;
    if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
    {
       p[0] = R;
       p[1] = G;
       p[2] = B;
    }
    else
    {
       p[0] = B;
       p[1] = G;
       p[2] = R;
    }
}



SDL_Rect set_rect(int x, int y, int w, int h)
{
	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	rect.w = w;
	rect.h = h;
	return rect;
}

SDL_Surface * prepare_text(int r, int b, int g, char *bufor, TTF_Font *font) {

   SDL_Color fgcolor;

   fgcolor.r = r;
   fgcolor.b = b;
   fgcolor.g = g;

   return TTF_RenderText_Solid(font, bufor,  fgcolor);
}

// rysuje obrazek
void draw (int x, int y, SDL_Surface * img, int trans) {

	SDL_Rect srcrect, destrect;

	// ustwaia przeźroczystość
	if (trans) SDL_SetColorKey(img,SDL_TRUE,getpixel(img,0,0));

	destrect = set_rect(x, y, img->w, img->h);
	srcrect = set_rect(0, 0, img->w, img->h);
	SDL_BlitSurface(img, &srcrect, game_screen, &destrect);
}

SDL_Surface * create_surface(int w, int h, int r, int g, int b) {

	SDL_Surface * surf;
	SDL_Rect dest;

	surf = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h,16,0,0,0,0);
	dest = set_rect(0, 0, w, h);
	SDL_FillRect(surf, &dest, SDL_MapRGB(surf->format, r, g, b));

	return surf;
}

void print_text(int r, int g, int b, char *c, int x, int y, TTF_Font * font) {

	SDL_Surface * info;
	//SDL_Surface * rubber;  // gumka

	// NOTE - gumka niepotrzebna gdyż odświeżane jest całe okno

	info = (SDL_Surface *) prepare_text(r, g, b, c, font);
	//rubber = create_surface(info->w, info->h, 65, 190, 215);

	//draw(x, y, rubber, 0);  // bez przezroczystości
	//SDL_FreeSurface(rubber);

	draw(x, y, info, 1);  // z przezroczystością
	SDL_FreeSurface(info);
}
/*
bool file_exist(){

	#if WINDOWS
	asm("nop");
	// kasuje zawarość
	//std::filesystem::path cwd = std::filesystem::current_path() / "lamertetris";
	//std::ofstream file(cwd.string());
	//return ( access( (char *) cwd.c_str(), F_OK ) != 0 );

	#elif LINUX
	
	char *file;
	char * FILE_NAME = "/.lamertetris";
	char * HOME = "HOME";

	if (!getenv(HOME)) exit (1);

	file=(char *)malloc(strlen(getenv(HOME))+strlen(FILE_NAME)+1);
	
	strcpy(file, getenv(HOME));
	strcat(file, FILE_NAME);
	printf("1");

	if (access(file, F_OK)==0){
		printf("PLIK istnieje %s",file);
		free(file);
		return true;
	}
	printf("PLIK nie istnieje %s",file);
	free(file);
	return false;
	#endif
}
*/
void save_hiscore(){

	#if WINDOWS
	std::string cfgString = std::to_string(hiscore);

	std::filesystem::path cwd = std::filesystem::current_path() / "lamertetris";
	std::ofstream file(cwd.string());
	if (file.is_open()) {
		file.write(cfgString.data(), cfgString.length());
	}
	file.close();
	#elif LINUX
	char *file;
	char * FILE_NAME = "/.lamertetris";
	char * HOME = "HOME";
	char bufor[48];
	if (!getenv(HOME)) exit (1);

	file=(char *)malloc(strlen(getenv(HOME))+strlen(FILE_NAME)+1);

	strcpy(file, getenv(HOME));
	strcat(file, FILE_NAME);
	
	sprintf(bufor,"%d",hiscore);

	FILE * f = fopen(file, "w");
	fwrite(bufor, sizeof(char), strlen(bufor), f);
	fclose(f);
	printf ("Zapisałem: %s\n", file);
	free(file);
	
	#endif
}

void load_hiscore(){

	#if WINDOWS

	std::filesystem::path cwd = std::filesystem::current_path() / "lamertetris";
	std::ifstream file(cwd.string());
	
	std::cout<<"PATH LOAD: "<<cwd.string()<<std::endl;
	
    std::string input;
    while (file >> input){
        hiscore=std::stoi(input);
	}
	//file.close();
	
	#elif LINUX
	char *file;
	char bufor[48];
	char * FILE_NAME = "/.lamertetris";
	char * HOME = "HOME";
	if (!getenv(HOME)) exit (1);

	//file=(char *)malloc(strlen(getenv(HOME)+strlen(FILE_NAME)+1));
	file=(char *)malloc(strlen(getenv(HOME))+strlen(FILE_NAME)+1);

	strcpy(file, getenv(HOME));
	strcat(file, FILE_NAME);

	FILE * f = fopen(file, "r");
	
	if (access(file, F_OK)==0){
		printf("PLIK istnieje %s",file);
		fscanf(f, "%s", bufor);
		fclose(f);
	}

	free(file);

	hiscore=atoi(bufor);
	#endif
	
// najlepszy=0;
}


bool game()
{
	int done=0;
	SDL_Rect source, dest;
	char txt_buf[] = "G A M E   O V  E R";
	char con[] = "Press SPACE to continue";
	char score_buf[128] = {'\0'};
	char hiscore_buf[128] = {'\0'};
	Field field;
	Szlam szlam;
	int color,n_element=0, figure=1, rate=1, tmp_element=0;
	SDL_Rect positionField, srcField;
	int left=0;
	int bottom=1;
	bool end = false;
	
	//if(file_exist())
		load_hiscore();
		
	positionField = set_rect(200, 0, 80, 0);
	srcField = set_rect(0, 0, 80, 80);
	
#if WINDOWS
	wallpaper = SDL_LoadBMP("img/wallpaper.bmp");
#elif LINUX
	wallpaper = SDL_LoadBMP("/usr/share/lamertetris/img/wallpaper.bmp");
#endif

	srand(time(NULL));
	figure = rand()%7;

	while(!done) {
		SDL_Event event;
		while (SDL_PollEvent (&event)) {

			switch (event.type) {

            case SDL_KEYDOWN:
               printf("Naciśnięto klawisz: %s\n", SDL_GetKeyName(event.key.keysym.sym));

               if (event.key.keysym.sym==SDLK_ESCAPE)
                  done = 1;

				if (event.key.keysym.sym==SDLK_SPACE) {
					if (end) {
						szlam.reset();
						end = false;
						score = 0;
						figure = rand()%7;
					}
				}
				if( event.type == SDL_KEYDOWN )
                {
                    //Select surfaces based on key press
					switch( event.key.keysym.sym )
                    {
                    case SDLK_UP:
						tmp_element=n_element;
						if(tmp_element<3) {
							tmp_element+=1;
						} else tmp_element=0;
					
						if(!szlam.collisionRotate(field.get(figure,tmp_element), tmp_element, 
							figure, positionField.x, positionField.y, field.getColor(figure), &bottom))
							if (field.checkEdgeRotate(field.get(figure, tmp_element), positionField.x)) {
								if(n_element<3) {
									n_element+=1;
								} else n_element=0;
							}
						break;
					case SDLK_LEFT:
						left=0;
						if(!szlam.collisionLeft(field.get(figure,n_element), n_element, 
							figure, positionField.x, positionField.y, field.getColor(figure), &bottom))
							if (field.checkEdgeLeft(field.get(figure, n_element), positionField.x))
								positionField.x-=20;
						
						break;
					case SDLK_RIGHT:
						if(!szlam.collisionRight(field.get(figure,n_element), n_element, 
							figure, positionField.x, positionField.y, field.getColor(figure), &bottom))
							if (field.checkEdgeRight(field.get(figure, n_element), positionField.x))
								positionField.x+=20;
							
						break;
					case SDLK_DOWN:
						rate=5;
						break;
					case SDLK_f:
						if(figure<NFIGURE-1) {
							figure+=1;
						} else figure=0;
						break;
					}
				}
				
				break;  // keydown
			case SDL_QUIT:
				if (score > hiscore) 
				{
					hiscore=score;
					save_hiscore();
				}
				SDL_DestroyTexture(tex_screen);
				SDL_DestroyRenderer(renderer);
    			SDL_DestroyWindow(gWindow);
				
               exit(0);
               break;

            case SDL_MOUSEMOTION:
				break;

			case SDL_MOUSEBUTTONDOWN:
				//if (!pauza) pauza=1;
				//else pauza=0;
				break;

			case SDL_MOUSEBUTTONUP:
				break;

            default:
            break;
			}
		}
		dest = set_rect(0, 0, game_screen->w, game_screen->h);
		SDL_FillRect(game_screen, &dest, SDL_MapRGB(game_screen->format, 255, 255, 255));
		draw(0, 0, wallpaper, 0);
		
			
		SDL_SetColorKey(field.get(figure,n_element),SDL_TRUE,
				getpixel(field.get(figure,n_element),79,79));
		// dla elementu wysuwanego
		if(srcField.y>0 && !end){
			SDL_BlitSurface( field.get(figure,n_element), &srcField, game_screen, &positionField );
			srcField.y--; srcField.h++;
			positionField.h++;

		} else  // dla elementu wysunietego
			SDL_BlitSurface( field.get(figure,n_element), &srcField, game_screen, &positionField );	
		
		szlam.draw();
		SDL_BlitSurface( szlam.get(), NULL, game_screen, NULL );
		
		
		if (end)
		{
			print_text(0, 0, 0, txt_buf,  40, 200, font_panel);
			print_text(255, 255, 255, con,  60, 250, font_score);
			if (score > hiscore) 
			{
				hiscore=score;
				save_hiscore();
			}
		}
		sprintf (score_buf, "%d", score);
		print_text(255, 255, 255, score_buf,  20, 1, font_score);

		sprintf (hiscore_buf, "%d", hiscore);
		print_text(255, 165, 0, hiscore_buf,  300, 1, font_score);
		
		if (bottom && szlam.collisionBottom(field.get(figure,n_element), n_element, 
				figure, positionField.x, positionField.y, field.getColor(figure))){
				if (!end){
                	positionField.y=0;
        			positionField.x=200;
					positionField.h=0;
					srcField.y=80;
					srcField.h=0;
      				figure = rand()%7;
          			rate=1;
          			score++;

          			// sprawdza czy równa powierzchnia
					int heightLevel=0, evenLevel = szlam.evenSurface(&heightLevel);
          			if (evenLevel) {
						szlam.cutSurface(evenLevel, heightLevel);
						score += 15*evenLevel;
        			}
				}
				
          		
        		
		}	
		else if (szlam.addBottom(field.get(figure,n_element), figure, n_element,
				field.getColor(figure),positionField.x,positionField.y )){
					positionField.y=0;
        			positionField.x=200;
					positionField.h=0;
					srcField.y=80;
					srcField.h=0;
      				figure = rand()%7;
          			rate=1;
          		
					score++;
          		// sprawdza czy równa powierzchnia
          		int heightLevel=0, evenLevel = szlam.evenSurface(&heightLevel);
          		if (evenLevel) {
					szlam.cutSurface(evenLevel, heightLevel);
					score += 15*evenLevel;
				} 

		}
		bottom = 1;
		if (szlam.highFull()) end = SDL_TRUE;
		// aż się wynurzy
		if (srcField.y==0 && !end)
				positionField.y+=rate;	
		//dest = set_rect(0, 0, game_screen->w, game_screen->h);
		source = set_rect(0, 0, game_screen->w, game_screen->h);
		tex_screen = SDL_CreateTextureFromSurface(renderer, game_screen);
		SDL_RenderClear ( renderer );
		SDL_RenderCopy(renderer, tex_screen,&source,&dest);
		SDL_RenderPresent(renderer);
		SDL_DestroyTexture(tex_screen);

	SDL_Delay(12);
	
	}
	return SDL_FALSE;
}

int main (int argc, char ** arg)
{
		
		
	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
	{
		printf ("Nie mozna zainicjowac SDL: %s\n", SDL_GetError ());
		exit (1);
	}
	atexit (SDL_Quit);
	
	/* inicjalizacja fontow */
	if(TTF_Init()==-1) {
		printf("TTF_Init: %s\n", TTF_GetError());
		exit(2);
	}
	atexit (TTF_Quit);


#if WINDOWS
#define FONT "Bevan.ttf"
#elif LINUX
#define FONT "/usr/share/lamertetris/Bevan.ttf"
#endif

   if((font_panel = TTF_OpenFont(FONT, SIZEFONT)) == NULL) {
       printf("TTF_OpenFont: %s\n", TTF_GetError());
       exit(1);
	}
	if((font_score = TTF_OpenFont(FONT, SCOREFONT)) == NULL) {
       printf("TTF_OpenFont: %s\n", TTF_GetError());
       exit(1);
	}

 // Ten sposób tworzenia okna i renderu gwarantuje że duszki nie będą przerywać
    gWindow = SDL_CreateWindow( "Lamer Tetris", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 400, 600, SDL_WINDOW_SHOWN );
	if( gWindow == NULL )
	{
		printf( "Window could not be created! SDL Error: %s\n", SDL_GetError() );
		return 1;
	}
    renderer = SDL_CreateRenderer( gWindow, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC );
	if( renderer == NULL )
	{
		printf( "Renderer could not be created! SDL Error: %s\n", SDL_GetError() );
		return 1;
	}
		// FIXME - Jednak prawda jest taka że nawet wywołanie SDL_GetWindowSurface przed SDL_CreateRenderer niewiele pomaga
		// gdyż obie funkcje się najprawdopodobniej nie lubią
		game_screen= SDL_GetWindowSurface( gWindow );
		if ( game_screen == NULL ) {
			printf("Cant create surface !\n");
			return -1;
		}
		asm("nop");
		
		SDL_SetRenderDrawColor ( renderer ,65, 190, 215 , 255 );
		SDL_RenderClear ( renderer );
	//}
	
#if LINUX
	SDL_Surface * icon = SDL_LoadBMP("/usr/share/lamertetris/img/lamertetris.bmp");
	SDL_SetWindowIcon(gWindow, icon);
#endif

	while(SDL_TRUE){
		if (game()==SDL_TRUE){
			
		}
		else break;
	}
	if (score > hiscore) 
	{
		hiscore=score;
		save_hiscore();
	}
	SDL_DestroyTexture(tex_screen);
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(gWindow);
	return 0;
}
